/* 			pdpcap.c 0.0.1 (beta) 					*/
/*		Jordi Sala poperbu@gmail.com 20100517				*/
/*										*/
/* --------------------------  pdpcap     ------------------------------------- */
/*                                                                              */
/* Is an External Object for pd that uses lipcap to capture and analizing	*/
/* network packets=> Network Sniffer                              		*/
/* sources and more info: http://musa.poperbu.net/puredata 	                */
/*                                                                              */
/* This program is free software; you can redistribute it and/or                */
/* modify it under the terms of the GNU General Public License                  */
/* as published by the Free Software Foundation; either version 2               */
/* of the License, or (at your option) any later version.                       */
/*                                                                              */
/* This program is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/* GNU General Public License for more details.                                 */
/*                                                                              */
/* You should have received a copy of the GNU General Public License            */
/* along with this program; if not, write to the Free Software                  */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.  */
/* ---------------------------------------------------------------------------- */
#include <pcap.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <net/ethernet.h>
#include <arpa/inet.h>
#include <pcap.h>
#include <ctype.h>

#include "m_pd.h"

/* default snap length (maximum bytes per packet to capture) */
#define SNAP_LEN 1518


#define IPPROTO_NVP 11

#define MAX_OUT_MESSAGE 65536L

EXTERN_STRUCT pcap_pkthdr;
#define t_header struct pcap_pkthdr

#define MAXFILTERARGS 10
#define MAXFILTERLENG MAXPDSTRING

static t_class *pdpcap_class;

typedef struct _pdpcap {
    t_object  x_obj;
    t_outlet *x_outlet1;        /*all line captured*/
    t_outlet *x_outlet2;        /*packet number*/
    t_outlet *x_outlet3;        /*protocol*/
    t_outlet *x_outlet4;        /*source IP address*/
    t_outlet *x_outlet5;        /*dest IP address*/
    t_outlet *x_outlet6;        /*src_port number*/
    t_outlet *x_outlet7;        /*dst_port number*/
    t_outlet *x_outlet8;        /*IP data size*/
    t_outlet *b_out;            /*end of capture bang*/
    t_clock *x_clock;

    int x_debug;
    int x_payloadon;            /*show on/off payload*/
    int x_num_packets;          /* number of packets to capture */
    int x_limited_capture;
    int x_previus_np;
    int x_connected;            /*pcap connected */
    int x_timeout;              /*timeou for pcap functions*/
    int x_delay;                /*delay for loop mode*/

    int x_count;                /*total packets processed*/
    int x_num_packets_count;    /*current packets processed*/
    int x_ip_count;             /*ip packets processed*/
    int x_tcp_count;            /*tcp packets processed*/
    int x_udp_count;            /*udp packets processed*/
    int x_icmp_count;           /*icmp packets processed*/
    int x_unk_count;            /*unknown packets processed*/
    int x_ip_count_t;             /*ip packets processed*/
    int x_tcp_count_t;            /*tcp packets processed*/
    int x_udp_count_t;            /*udp packets processed*/
    int x_icmp_count_t;           /*icmp packets processed*/
    int x_unk_count_t;            /*unknown packets processed*/

    t_float x_loop;             /*loop mode on/off inlet2*/

    //pcap
    bpf_u_int32 x_mask;			    /* subnet mask */
    bpf_u_int32 x_net;			    /* ip */
    char x_errbuf[PCAP_ERRBUF_SIZE];	/* error buffer */
    struct bpf_program x_fp;		/* compiled filter program (expression) */
    pcap_t *x_handle;			    /* packet capture handle */
    char *x_dev;		            /* capture device name */
    const u_char *x_packet;         /*captured packet*/
    struct pcap_pkthdr *x_header;   /*packet header*/
    char *x_filter_exp;		        /* filter expression [3] */
} t_pdpcap;


void pdpcap_tick(t_pdpcap *x);
void pdpcap_got_packet(t_pdpcap *x);

void pdpcap_connect(t_pdpcap *x)
{
    t_float c;
    if (x->x_num_packets == 0){x->x_num_packets=10;}

    x->x_dev = pcap_lookupdev(x->x_errbuf);

    if (x->x_dev == NULL) {
            error("pdpcap: Couldn't find default device.");
	}
	else{
             post("pdpcap: Detected Device: %s", x->x_dev);
            /* get network number and maskmask associated with capture device */
            if (pcap_lookupnet(x->x_dev, &x->x_net, &x->x_mask, x->x_errbuf) == -1) {
                error("pdpcap: Couldn't get netmask for device %s: %s",x->x_dev, x->x_errbuf);
                x->x_net= 0;
                x->x_mask = 0;
                x->x_connected=0;

            }
            else{
                /* print capture info */
                 post("pdpcap: Device: %s", x->x_dev);
                 if (x->x_debug)post("pdpcap: Number of packets: %d", x->x_num_packets);
                 if (x->x_debug)post("pdpcap: Filter expression: %s", x->x_filter_exp);

                /* open capture device */
                //x->x_handle = pcap_open_live(x->x_dev, SNAP_LEN, 1, 1000, x->x_errbuf);
                x->x_handle = pcap_open_live(x->x_dev, SNAP_LEN, 1, x->x_timeout, x->x_errbuf);
                if (x->x_handle == NULL) {
                    error("pdpcap: Couldn't open device %s: %s", x->x_dev, x->x_errbuf);
                    x->x_connected=0;
                }
                else{
                     post("pdpcap: device %s: %s opened", x->x_dev, x->x_errbuf);
                    /* make sure we're capturing on an Ethernet device [2] */
                    /*if (pcap_setnonblock(x->x_handle, 1, x->x_errbuf) == -1) {
                        post("pdpcap: Failed to set NonBlocking mode.");
                    }else{
                        post("pdpcap: NonBlcocking mode.");
                    }*/
                    if (pcap_datalink(x->x_handle) != DLT_EN10MB) {
                        error("pdpcap: %s is not an Ethernet", x->x_dev);
                        x->x_connected=0;
                    }
                    else{
                        post("pdpcap: %s is an Ethernet->ok", x->x_dev);
                        /* compile the filter expression */
                        if (pcap_compile(x->x_handle, &x->x_fp, x->x_filter_exp, 0, x->x_net ) == -1) {
                            error("pdpcap: Couldn't parse filter %s: %s", x->x_filter_exp, pcap_geterr(x->x_handle));
                            x->x_connected=0;
                        }
                        else{
                            post("pdpcap: Filter Parsed %s: %s -> Ok", x->x_filter_exp, pcap_geterr(x->x_handle));
                            /* apply the compiled filter */
                            if (pcap_setfilter(x->x_handle, &x->x_fp) == -1) {
                                error("pdpcap: Couldn't install filter %s: %s", x->x_filter_exp, pcap_geterr(x->x_handle));
                                x->x_connected=0;
                            }
                            else{
                                post("pdpcap: Filter Installed  %s: %s-> Ok", x->x_filter_exp, pcap_geterr(x->x_handle));
                                x->x_connected=1;

                            }
                        }
                    }
                }

            }
	}
	c=x->x_connected;
	outlet_float(x->x_outlet1,c);

}
void pdpcap_disconnect(t_pdpcap *x){
    t_float c;
    if (x->x_connected == 0){
        post("pdpcap: device was not connected NOT connected.");
    }else{
        post("pdpcap: OK device disconnected.");
        pcap_freecode(&x->x_fp);
        pcap_close(x->x_handle);
       // pcap_freealldevs(x->x_dev);
        x->x_connected=0;
        c=x->x_connected;
        outlet_float(x->x_outlet1,c);
    }
}

void pdpcap_capture(t_pdpcap *x){
    int i;
    int ok;
    int next_res;
    next_res=0;
    ok=0;

    const u_char *packet;
    struct pcap_pkthdr *header;
    header=x->x_header;
    if (x->x_connected == 0){
        error("pdpcap: NOT connected.");
    }else{
        if (x->x_num_packets_count == 0){
                x->x_ip_count=0;
                x->x_tcp_count=0;
                x->x_udp_count=0;
                x->x_icmp_count=0;
                x->x_unk_count=0;
        }
        if (x->x_loop==0){
            if(x->x_num_packets_count < x->x_num_packets){
                if (x->x_debug)post("pdpcap: capturing %d of %d packets", x->x_num_packets_count, x->x_num_packets);
                if (x->x_num_packets_count==1)x->x_limited_capture=1;
                ok=1;
            }else{
                ok=0;
            }
        }else{
            ok=1;
            if (x->x_num_packets_count==1)
            {
                x->x_previus_np=x->x_num_packets;
                x->x_num_packets=1;
                x->x_limited_capture=0;
            }
        }
        if (ok==1){
            next_res=pcap_next_ex(x->x_handle, &x->x_header, &x->x_packet);

            if (next_res == 0){
                if (x->x_debug)post("pdpcap: waiting for paquets...");
            }
            if (next_res == 1){
                pdpcap_got_packet(x);
                outlet_bang(x->b_out);
            }
            if(next_res==-1){
                error("pdpcap: ERROR %s",x->x_errbuf);
            }
            //if(x->x_loop == 1){
            clock_delay(x->x_clock,x->x_delay);
            //}
            pcap_freecode(&x->x_fp);

            //if (x->x_debug)post("pdpcap: Capture complete.");
        }else{
            post("pdpcap: Capture finished");
            post("pdpcap:------- TOTAL num pkt processed: %d", x->x_count);
            post("pdpcap:                             ip: %d", x->x_ip_count_t);
            post("pdpcap:                            tcp: %d", x->x_tcp_count_t);
            post("pdpcap:                            udp: %d", x->x_udp_count_t);
            post("pdpcap:                           icmp: %d", x->x_icmp_count_t);
            post("pdpcap:                         others: %d", x->x_unk_count_t);
            x->x_ip_count_t=x->x_ip_count_t+x->x_ip_count;
            x->x_tcp_count_t=x->x_tcp_count_t+x->x_tcp_count;
            x->x_udp_count_t=x->x_udp_count_t+x->x_udp_count;
            x->x_icmp_count_t=x->x_icmp_count_t+x->x_icmp_count;
            x->x_unk_count_t=x->x_unk_count_t+x->x_unk_count;

            if (x->x_limited_capture=1){
                post("pdpcap: Last Capture num pkt processed: %d", x->x_num_packets_count);
                post("pdpcap:                             ip: %d", x->x_ip_count);
                post("pdpcap:                            tcp: %d", x->x_tcp_count);
                post("pdpcap:                            udp: %d", x->x_udp_count);
                post("pdpcap:                           icmp: %d", x->x_icmp_count);
                post("pdpcap:                         others: %d", x->x_unk_count);

                x->x_num_packets_count=0;
            }else{
                x->x_num_packets=x->x_previus_np;
            }
        }
    }
}

void pdpcap_got_packet(t_pdpcap *x)
{
	x->x_count++;
	//if (x->x_loop == 0){
    x->x_num_packets_count++;
	//}
	struct ether_header *ether;
	struct iphdr *ip;
	int len;
	int packet_counter;
	struct udphdr *udp;
	struct tcphdr *tcp;
    char src_ip[20], dest_ip[20];
    t_float sp;
    t_float dp;
    t_float ln;

    sp=0;
    dp=0;

	packet_counter=x->x_count;

    t_float f = x->x_count;
    outlet_float(x->x_outlet2, f);

	if (x->x_header->caplen != x->x_header->len) {
		error("pdpcap: %d != %d!!! Don't have complete packet. Skipping.\n",
			x->x_header->caplen, x->x_header->len);
		goto exit;
	}

	ether = (struct ether_header *) x->x_packet;

	if (ntohs(ether->ether_type) != ETHERTYPE_IP) {
		error("pdpcap: Packet %d: protocol is not IP. Skipping.", packet_counter);
		goto exit;
	}

	ip = (struct iphdr *) (x->x_packet + sizeof(struct ether_header) );

    len = ntohs(ip->tot_len) - sizeof(struct iphdr);
	ln =len;
    outlet_float(x->x_outlet8, ln);

    /* Get source and destination addresses */
	strcpy(src_ip, inet_ntoa( *(struct in_addr *) &ip->saddr) );
	strcpy(dest_ip, inet_ntoa( *(struct in_addr *) &ip->daddr) );

     if (x->x_debug){
        post("-------------Packet number: %d-------------------", packet_counter);
        post("Length of IP data: %d bytes", len);
        post("%s -> %s", src_ip, dest_ip);
        post("Length of IP data: %d", ntohs(ip->tot_len) -
				sizeof(struct iphdr) );
     }

    outlet_symbol(x->x_outlet4, gensym(src_ip));
    outlet_symbol(x->x_outlet5, gensym(dest_ip));

    switch (ip->protocol) {
		case IPPROTO_ICMP:
                outlet_symbol(x->x_outlet3, gensym("ICMP"));
                outlet_float(x->x_outlet6, sp);
                outlet_float(x->x_outlet7, dp);
                x->x_icmp_count=x->x_icmp_count++;
                if (x->x_debug){
                    post("IP protocol is ICMP");
                    post("-------------------------------------------------");

                }
                break;

		case IPPROTO_TCP:
                outlet_symbol(x->x_outlet3, gensym("TCP"));
                tcp = (struct tcphdr *) ( (char *) ip + sizeof(struct iphdr) );

                sp = ntohs(tcp->source);
                outlet_float(x->x_outlet6, sp);

                dp = ntohs(tcp->dest);
                outlet_float(x->x_outlet7, dp);

                x->x_tcp_count=x->x_tcp_count++;
                if (x->x_debug){
                    post("IP protocol is TCP");
                    post("Source port: %d, destination port %d",
                        ntohs(tcp->source), ntohs(tcp->dest) );
                    post("-------------------------------------------------");

                }
                break;

		case IPPROTO_UDP:
                outlet_symbol(x->x_outlet3, gensym("UDP"));
                udp = (struct udphdr *) ( (char *) ip + sizeof(struct iphdr) );

                sp = ntohs(udp->source);
                outlet_float(x->x_outlet6, sp);

                dp = ntohs(udp->dest);
                outlet_float(x->x_outlet7, dp);

                x->x_udp_count=x->x_udp_count++;

                if (x->x_debug){
                    post("IP protocol is UDP");
                    post("Source port: %d, destination port %d",
                        ntohs(udp->source), ntohs(udp->dest) );
                    post("-------------------------------------------------");

                }
                break;

		default:
                outlet_symbol(x->x_outlet3, gensym("UNKNOWN"));
                outlet_float(x->x_outlet6, sp);
                outlet_float(x->x_outlet7, dp);
                x->x_unk_count=x->x_unk_count++;
                if (x->x_debug){
                    post("Unknown IP protocol %d.", ip->protocol);
                    post("-------------------------------------------------");

                }
	}
    return;

    exit:
        if (x->x_debug){
            post("pdpcap: Some problem in packet sizes.");
        }

}

void pdpcap_numpackets_set(t_pdpcap *x, t_floatarg f)
{
    x->x_num_packets=f;
    post("pdpcap: Number of packets:%d",x->x_num_packets);
}

void pdpcap_loop_set(t_pdpcap *x, t_floatarg f)
{
    x->x_loop=f;
    x->x_num_packets=1;
    post("pdpcap: Loop set to:%d",x->x_loop);
    post("pdpcap: Number of packets (by default in loop mode):%d",x->x_num_packets);
}


void pdpcap_timeout_set(t_pdpcap *x, t_floatarg f)
{
    x->x_timeout=f;
    post("pdpcap: Timeout set to:%d",x->x_timeout);
}

void pdpcap_filter_set(t_pdpcap *x, t_symbol *s, int argc, t_atom *argv)
{
    typedef char cstring[MAXFILTERLENG];
    int i;
    cstring*buf=(cstring*)getbytes(MAXFILTERARGS*sizeof(cstring));
    cstring buffer;
    for(i=0; i<MAXFILTERARGS; i++)*buf[i]=0;
    for (i=0; i<argc; i++) {
		atom_string(argv+i, buf[i], MAXFILTERLENG);
        if (i<(argc-1)){
            strcat(buf[i]," ");
        }
	}
    sprintf(x->x_filter_exp,"%s%s%s%s%s%s%s%s%s%s",buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7], buf[8], buf[9]);
    freebytes(buf, MAXFILTERARGS*sizeof(cstring));
	s=gensym(buffer);
	post("pdpcap: New pcap Filter %s.", x->x_filter_exp);

}

void pdpcap_delay_loop_set(t_pdpcap *x, t_floatarg f)
{
    x->x_delay=f;
    post("pdpcap: Delay set to:%d",x->x_delay);
}

void pdpcap_debug_set(t_pdpcap *x, t_floatarg f)
{
    if (f==0 || f==1){
        x->x_debug=f;
        post("pdpcap: Debug set to:%d",x->x_debug);
    }
    else{
        error("pdpcap: Wrong value for debug (0 or 1)");
    }
}

void pdpcap_payloadon_set(t_pdpcap *x, t_floatarg f)
{
    if (f==0 || f==1){
        x->x_payloadon=f;
        post("pdpcap: Payload set to:%d",x->x_payloadon);
    }
    else{
        error("pdpcap: Wrong value for payload (0 or 1)");
    }
}

void pdpcap_info(t_pdpcap *x)
{
    post("pdpcap: connected: %d", x->x_connected);
    post("pdpcap: debug: %d", x->x_debug);
    post("pdpcap: num_packets to capture: %d", x->x_num_packets);
    post("pdpcap: timeout: %d", x->x_timeout);
    post("pdpcap: delay: %d", x->x_delay);
    post("pdpcap: Total num pkt processed: %d", x->x_count);
    post("pdpcap: Last capture num pkt processed: %d", x->x_num_packets_count);
    post("pdpcap:------- TOTAL num pkt processed: %d", x->x_count);
    post("pdpcap:                             ip: %d", x->x_ip_count_t);
    post("pdpcap:                            tcp: %d", x->x_tcp_count_t);
    post("pdpcap:                            udp: %d", x->x_udp_count_t);
    post("pdpcap:                           icmp: %d", x->x_icmp_count_t);
    post("pdpcap:                         others: %d", x->x_unk_count_t);
}

static void pdpcap_delay_stop(t_pdpcap *x)
{
  clock_unset(x->x_clock);
}
void pdpcap_tick(t_pdpcap *x)
{
  pdpcap_capture(x);
}

void pdpcap_reset(t_pdpcap *x)
{
    x->x_num_packets=1;
    x->x_num_packets_count=0;
    x->x_debug=0;
    x->x_connected=0;
    x->x_count=0;
    x->x_timeout=500;
    x->x_delay=100;

    x->x_ip_count_t=0;
    x->x_tcp_count_t=0;
    x->x_udp_count_t=0;
    x->x_icmp_count_t=0;
    x->x_unk_count_t=0;

  pdpcap_info(x);
}
static void pdpcap_free(t_pdpcap *x)
{
  clock_free(x->x_clock);
}

void pdpcap_bang(t_pdpcap *x)
{
    post("bang");
}

void *pdpcap_new(void)
{
    t_pdpcap *x = (t_pdpcap *)pd_new(pdpcap_class);
    floatinlet_new(&x->x_obj, &x->x_loop);
    x->x_outlet1 = outlet_new(&x->x_obj, &s_float);	    /* all line */
    x->x_outlet2 = outlet_new(&x->x_obj, &s_float);           /*packet number*/
    x->x_outlet3 = outlet_new(&x->x_obj, &s_float);           /*protocol*/
    x->x_outlet4 = outlet_new(&x->x_obj, &s_symbol);          /*source IP address*/
    x->x_outlet5 = outlet_new(&x->x_obj, &s_symbol);          /*dest IP address*/
    x->x_outlet6 = outlet_new(&x->x_obj, &s_float);           /*src_port number*/
    x->x_outlet7 = outlet_new(&x->x_obj, &s_float);           /*dst_port number*/
    x->x_outlet8 = outlet_new(&x->x_obj, &s_float);           /*IP data size*/


    x->b_out = outlet_new(&x->x_obj, &s_bang);                /*end of capture bang*/
    x->x_clock = clock_new(x, (t_method)pdpcap_tick);


    x->x_filter_exp = (char *)getbytes(MAXFILTERLENG * sizeof(char));

    x->x_num_packets=1;
    x->x_num_packets_count=0;
    x->x_debug=0;
    x->x_connected=0;
    x->x_count=0;
    x->x_timeout=500;
    x->x_delay=100;
    x->x_ip_count_t=0;
    x->x_tcp_count_t=0;
    x->x_udp_count_t=0;
    x->x_icmp_count_t=0;
    x->x_unk_count_t=0;

return (void *)x;
}

void pdpcap_setup(void) {
  pdpcap_class = class_new(gensym("pdpcap"),
        (t_newmethod)pdpcap_new,
        0, sizeof(t_pdpcap),
        CLASS_DEFAULT, 0);
    class_addbang(pdpcap_class, pdpcap_bang);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_connect, gensym("connect"),0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_disconnect, gensym("disconnect"),0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_capture, gensym("capture"),0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_numpackets_set, gensym("packets"),
        A_DEFFLOAT, 0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_timeout_set, gensym("timeout"),
        A_DEFFLOAT, 0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_delay_loop_set, gensym("delay"),
        A_DEFFLOAT, 0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_debug_set, gensym("debug"),
        A_DEFFLOAT, 0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_reset, gensym("reset"),
        A_DEFFLOAT, 0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_info, gensym("info"),
        A_DEFFLOAT, 0);
    class_addmethod(pdpcap_class,
        (t_method)pdpcap_filter_set, gensym("filter"),
        A_GIMME,0);


    class_sethelpsymbol(pdpcap_class, gensym("pdpcap_help"));
}
